import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly


def all_trace(da):
    '''
    Plot all the trace using plotly. The number of trace should not be to high.
    '''
    # Determine the number of rows (channels) in da
    num_rows = da.shape[1]

    # Create a subplot with shared x-axes and `num_rows` subplots
    fig = make_subplots(rows=num_rows, cols=1, shared_xaxes=True, subplot_titles=[f'Distance {da.distance[i].values} m' for i in range(num_rows)])

    # Loop through each row to create a plot
    for i in range(num_rows):
        fig.add_trace(go.Scatter(
            x=da['time'].values,
            y=da[:, i].values,
            mode='lines',
            name=f'Distance {i} m'
        ), row=i + 1, col=1)

        # Hide the x-axis label except for the last plot
        if i < num_rows - 1:
            fig.update_xaxes(showticklabels=False, row=i + 1, col=1)

    # Update layout with a tight layout and margin adjustments
    fig.update_layout(
        title='Synchronized Strain Rate Plots',
        xaxis_title='Time',
        yaxis_title='Strain Rate',
        height=300 * num_rows,  # Adjust height dynamically based on the number of plots
        margin=dict(l=50, r=50, t=50, b=50)  # Adjust margins for a tighter layout
    )

    return fig


import xarray as xr

def std_local(da,nt=6,nx=4,ax_t='time',ax_x='distance'):
    '''
    Compute local standard deviation for a DataArray

    :param da: Variable on which compute the local standard deviation
    :type da: xr.DataArray
    :param nt: number of element in ax_t direction (default: 6)
    :type nt: int
    :param nx: number of element in ax_x direction (default: 4)
    :type nx: int
    :param ax_t: ax_t axis (default: 'time')
    :type ax_t: string
    :param ax_x: ax_x axis (default: 'distance')
    :type ax_x: string
    '''

    return da.rolling({ax_t:nt,ax_x:nx}).std(ax_t)

def strain_rate(da,ax_d='distance'):
    '''
    Compute strain rate from velocity mesurement

    :param da: Variable on which compute differentiate
    :type da: xr.DataArray
    :param ax_d: axis for differentiate (default: 'distance')
    :type ax_d: string
    '''

    return da.differentiate(ax_d)


import warnings
import xarray as xr
import numpy as np
from scipy.signal import butter, sosfiltfilt

class DecimationWarning(UserWarning):
    """Custom warning for decimation issues."""
    pass

def decimate(da, q_int, axis='time'):
    """
    Decimate an xarray.DataArray or Dataset along a specified axis by a given interval.

    Parameters:
    da : xarray.DataArray or xarray.Dataset
        The input data to be decimated.
    q_int : int
        The decimation interval. Must be a positive integer.
    axis : str, optional
        The dimension name along which to decimate. Default is 'time'.

    Returns:
    xarray.DataArray or xarray.Dataset
        The decimated data.

    Raises:
    ValueError:
        If q_int is not a positive integer.
    """
    if not isinstance(q_int, int) or q_int <= 0:
        raise ValueError("q_int must be a positive integer.")

    if axis not in da.dims:
        warnings.warn(f"{axis} is not in the dimensions of the DataArray or Dataset.", 
                      DecimationWarning, stacklevel=2)
        return da
    
    # Perform decimation along the specified axis
    return da.isel(**{axis: slice(0, None, q_int)})

def bandpass(da, lowcut, highcut, order=4, axis='time'):
    """
    Apply a bandpass filter to an xarray.DataArray along a specified axis lazily, compatible with Dask.
    
    Parameters:
    da : xarray.DataArray
        Input DataArray to be filtered. Can be Dask-backed for lazy evaluation.
    lowcut : float
        Low cutoff frequency for the bandpass filter.
    highcut : float
        High cutoff frequency for the bandpass filter.
    order : int, optional
        Order of the Butterworth filter. Default is 4.
    axis : str, optional
        The dimension along which to apply the filter. Default is 'time'.

    Returns:
    xarray.DataArray
        The bandpass-filtered DataArray, computed lazily if Dask-backed.
    """
    if axis not in da.dims:
        raise ValueError(f"Axis '{axis}' is not in the dimensions of the DataArray: {list(da.dims)}")

    # Determine sampling frequency
    dT = (da[axis][1] - da[axis][0]).values.astype(float)
    fs = 1 / (dT * 1e-9)  # Convert nanoseconds to seconds for sampling rate

    # Design the Butterworth bandpass filter using SOS (more stable than b/a coefficients)
    sos = butter(order, [lowcut, highcut], fs=fs, btype='bandpass', output='sos')

    # Get the axis index
    axis_index = list(da.dims).index(axis)

    # Apply the filter using xr.apply_ufunc
    def apply_filter(data):
        return sosfiltfilt(sos, data, axis=axis_index)

    da_bp = xr.apply_ufunc(
        apply_filter,                # Function to apply
        da,                          # Input array
        input_core_dims=[[axis]],    # Core dimension to apply the function on
        output_core_dims=[[axis]],   # Core dimension for the output
        vectorize=True,              # Apply the function element-wise across chunks
        dask="parallelized",         # Enable Dask parallelization
        output_dtypes=[da.dtype],    # Specify output data type
        dask_gufunc_kwargs={"allow_rechunk": True},  # Allow rechunking along the core dimension
    )

    return da_bp.T



import h5py
import numpy as np
from xdas import DataArray
from xdas.virtual import VirtualSource
from datetime import datetime,timezone
import xarray as xr
from natsort import natsorted
import glob
import pandas as pd

def read_Terra15(fname,timezone=timezone.utc):
  '''
  Open file from Terra15 version 6 using xdas loader

  :param timezone: default (timezone.utc)
  :type timezone: timezone 
  '''

  with h5py.File(fname, "r") as file:
    ti=np.datetime64(datetime.fromtimestamp(file['data_product']['gps_time'][0],tz=timezone)).astype('datetime64[ms]')
    tf=np.datetime64(datetime.fromtimestamp(file['data_product']['gps_time'][-1],tz=timezone)).astype('datetime64[ms]')
    d0 = file.attrs['sensing_range_start']
    dx = file.attrs['dx']
    data = VirtualSource(file['data_product']['data'])
  nt, nd = data.shape
  t = {"tie_indices": [0, nt - 1], "tie_values": [ti, tf]}
  d = {"tie_indices": [0, nd - 1], "tie_values": [d0, d0+ (nd - 1) * dx]}
  return DataArray(data, {"time": t, "distance": d})


def dask_Terra15(fname,timezone=timezone.utc,**kwargs):
  '''
  Open data using xarray dask 

  :param timezone: default (timezone.utc)
  :type timezone: timezone 
  '''

  ds=xr.open_mfdataset(fname,engine='h5netcdf',group='/data_product',phony_dims='sort',concat_dim='phony_dim_0',combine='nested',**kwargs)

  ds=ds.rename({'phony_dim_0': 'time'})


  if fname is list:
    da=xr.open_dataset(fname[0])
  else: 
    da=xr.open_dataset(natsorted(glob.glob(fname))[0])

  d0 = da.attrs['sensing_range_start']
  dx = da.attrs['dx']

  nt, nd = ds.data.shape

  d = np.linspace(d0,d0+(nd-1)*dx,nd)

  ds=ds.rename({'phony_dim_1': 'distance'})
  ds['distance']=d


  # Convert timestamp in numpy datetime[ns]
  pdt0=ds['gps_time'].to_pandas()
  dt=da.attrs['dt_computer']
  pdt0 = pdt0.astype(float)
  pdt = np.linspace(pdt0[0],pdt0[0]+(nt-1)*dt,nt)
  
  #vectorized_function = np.vectorize(lambda x: datetime.fromtimestamp(x, tz=timezone))
  #pdt = vectorized_function(pdt)

  #pdt = pdt.apply(lambda x: datetime.fromtimestamp(x,tz=timezone))
  #ds['time']= [pd.Timestamp(datetime.utcfromtimestamp(ts).replace(tzinfo=timezone.utc)).to_pydatetime() for ts in pdt]
  
  ti=datetime.fromtimestamp(pdt[0],tz=timezone.utc)
  tf=datetime.fromtimestamp(pdt[-1],tz=timezone.utc)
  ds['time']=pd.date_range(ti,tf,nt).values



  # Copy attributs
  ds.attrs=da.attrs
  ds.attrs['nt']=nt

  # drop unwanted variable
  ds=ds.drop_vars('posix_time')
  ds=ds.drop_vars('gps_time')
  ds=ds.data.rename('velocity')

  return ds

def dask_febus_one_file(fname, timezone=timezone.utc, **kwargs):
    """
    Open and process Febus DAS data using xarray with dask.

    :param fname: File name or pattern to load with xarray.
    :param timezone: Timezone for datetime conversion (default: UTC).
    :type timezone: timezone object
    :param kwargs: Additional arguments for xarray.open_mfdataset.
    :return: xarray DataArray with time and distance coordinates.
    """

    # Open file with h5py to extract metadata
    with h5py.File(fname, 'r') as fh5f:
        # Extract source and zone attributes for metadata
        src_attr = fh5f['localhost.localdomain']['Source1'].attrs
        zne_attr = fh5f['localhost.localdomain']['Source1']['Zone1'].attrs

        # Extract key parameters from attributes
        prf = zne_attr['PulseRateFreq'] / 1000  # Pulse rate frequency in kHz
        block_rate = zne_attr['BlockRate'] / 1000  # Block rate in kHz
        sampling_res = src_attr['SamplingRes']
        derivation_time = zne_attr['DerivationTime']
        origin = zne_attr['Origin']
        spacing = zne_attr['Spacing']
        gauge_length = zne_attr['GaugeLength']
        overlap = zne_attr['BlockOverlap'][0]  # Overlap percentage

        # Calculate spatial and temporal resolutions
        dt = spacing[1] * 1.e-3  # Temporal resolution in seconds
        dx = spacing[0]          # Spatial resolution in meters

        # Determine data shape and read block times
        channel = list(fh5f['localhost.localdomain']['Source1']['Zone1'].keys())[0]
        node = fh5f['localhost.localdomain']['Source1']['Zone1'][channel]
        block_times = fh5f['localhost.localdomain']['Source1']['time'][:]
        nb_block, block_time_size, block_space_size = node.shape

    # Calculate block-related parameters
    block_time_useful_size = int(100 / (100 + overlap) * block_time_size)
    block_time_overlap_size = int((block_time_size - block_time_useful_size) / 2)
    block_time_offset = (block_time_useful_size / 2) * dt

    # Determine start and end times for useful data in blocks
    block_time_useful_start_time = block_times - block_time_offset
    block_time_useful_end_time = block_time_useful_start_time + block_time_useful_size * dt

    # Open data with xarray using h5netcdf engine
    da = xr.open_mfdataset(
        fname,
        engine='h5netcdf',
        group='localhost.localdomain/Source1/Zone1/',
        phony_dims='sort',
        combine='nested',
        **kwargs
    )

    # Extract the channel data and calculate offsets
    t1, t2, d1 = da[channel].shape
    offset = int((t2 - block_time_useful_size) / 2)

    # Stack dimensions for easier manipulation
    da_stacked = da[channel][:, offset:block_time_useful_size + offset].stack(
        new_dim=('phony_dim_1', 'phony_dim_2')
    )

    # Generate time and distance coordinates
    time = block_time_useful_start_time[0] + np.arange(da_stacked.shape[1]) * dt
    ti = datetime.fromtimestamp(time[0], tz=timezone)
    tf = datetime.fromtimestamp(time[-1], tz=timezone)
    distance = np.arange(d1) * dx

    # Create new time and distance dimensions
    da_stacked = da_stacked.drop_vars(['new_dim', 'phony_dim_1', 'phony_dim_2'])
    da_stacked['new_dim'] = pd.date_range(ti, tf, periods=da_stacked.shape[1]).values
    da_stacked['phony_dim_3'] = distance

    # Rename dimensions for clarity
    da_stacked = da_stacked.rename({'phony_dim_3': 'distance', 'new_dim': 'time'})

    # Add attribut information. It might be nice to have them consistent between terra15 and febus
    da_stacked.attrs['dt']=dt
    da_stacked.attrs['dx']=dx
    da_stacked.attrs['overlap']=overlap

    # Transpose to have 'time' and 'distance' as primary coordinates
    return da_stacked.transpose()

def dask_febus(filenames, timezone=timezone.utc, concat_dim='time', **kwargs):
    """
    Load and combine multiple Febus DAS files using dask_febus_one_file.

    :param filenames: List of file paths to process.
    :param timezone: Timezone for datetime conversion (default: UTC).
    :type timezone: timezone object
    :param concat_dim: Dimension along which to concatenate the data (default: 'time').
    :param kwargs: Additional arguments to pass to dask_febus_one_file.
    :return: Combined xarray.DataArray object.
    """
    data_arrays = []

    for fname in filenames:
        try:
            # Load each file using dask_febus_one_file
            da = dask_febus_one_file(fname, timezone=timezone, **kwargs)
            data_arrays.append(da)
        except Exception as e:
            print(f"Error processing file {fname}: {e}")

    # Concatenate all data along the specified dimension
    combined_da = xr.concat(data_arrays, dim=concat_dim)

    return combined_da


import xarray as xr
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

# Function to process and apply optic length and GPS data on data

def optic_length(da,metadata_optic,dist='distance'):
    '''
    Compute optic length from metadata
    '''

    meta_op=pd.read_csv(metadata_optic)
    meta_op['optic_distance']=np.cumsum((meta_op.Cable_Start-meta_op.Cable_End).abs()*(1+meta_op.Loop))

    # Create 'name' DataArray initialized with '-'
    da_res = xr.DataArray([None] * len(da[dist]), dims=[dist])

    i_start = 0
    i=0
    for i_end in meta_op.optic_distance:
        # Create a boolean mask for the range between i_start and i_end
        npa = ((da[dist] > i_start) & (da[dist] < i_end)).values
        
        # Check if there are any True values in the mask
        if np.any(npa):
            # Ensure you assign a single value (string) to all True positions
            da_res[np.where(npa==True)] = str(meta_op.Name[i] + '-' + meta_op.Type[i])  # Assigning the same single string to all True positions
        
        # Update i_start to the current i_end for the next iteration
        i_start = i_end
        i+=1

    return da_res

def optic_length_plot(da):
    '''
    Plot optic length data
    '''

    # Assuming your DataArray is named optic_length_distance
    # For example:
    optic_length_distance = da

    # Convert xarray DataArray to pandas DataFrame
    df = optic_length_distance.to_dataframe(name='name').reset_index()

    # Drop rows where 'name' is None
    df = df.dropna(subset=['name'])

    # Group by name and calculate the min and max distances
    grouped = df.groupby('name')['distance'].agg(['min', 'max']).reset_index()

    # Generate a unique color for each name using the new method
    colors = plt.get_cmap('tab20', len(grouped))  # You can use other colormaps as well

    # Plot the horizontal bars on a fixed y-coordinate
    plt.figure(figsize=(12, 3))

    # Initialize the y-coordinate for bars
    base_y_position = 0

    # List to store transition points for xticks and their colors
    xticks = []
    xtick_colors = {}

    # Plot bars with different colors
    for idx, row in grouped.iterrows():
        # Plot a horizontal bar
        bar_length = row['max'] - row['min']
        bar_color = colors(idx)
        plt.barh(base_y_position, bar_length, left=row['min'], 
                color=bar_color, edgecolor='black')
        
        # Calculate the midpoint of the bar
        midpoint = (row['max'] + row['min']) / 2
        
        # Add vertical text in the middle of the bar if the bar length is greater than 50
        if bar_length > 50:
            plt.text(midpoint, base_y_position, row['name'],
                    va='center', ha='center', color='black', fontsize=10, rotation=90)
        
        # Add min and max values to xticks if the bar length is greater than 50
        if bar_length > 50:
            xticks.extend([row['min'], row['max']])
            xtick_colors[row['min']] = bar_color
            xtick_colors[row['max']] = bar_color

    # Remove duplicate xticks and sort them
    xticks = sorted(set(xticks))

    # Plot customization
    plt.yticks([base_y_position], [''])  # Hide y-axis label for the single line
    plt.xlabel('Distance')

    # Set xticks at each transition
    plt.xticks(xticks)

    # Customize tick appearance
    ax = plt.gca()

    # Define the offsets for top and bottom positioning
    bottom_offset = 0
    top_offset = 1.05

    # Apply color to xtick labels
    for i, tick in enumerate(ax.get_xticks()):
        tick_label = ax.get_xticklabels()[i]
        tick_value = tick
        color = xtick_colors.get(tick_value, 'black')  # Default to black if not found
        
        # Set the color of the tick label
        tick_label.set_color(color)
        
        # Position the tick labels
        if (i // 2) % 2 == 0:
            # Even pairs: bottom
            tick_label.set_verticalalignment('top')
            tick_label.set_horizontalalignment('right')
            tick_label.set_y(bottom_offset)
        else:
            # Odd pairs: top
            tick_label.set_verticalalignment('bottom')
            tick_label.set_horizontalalignment('left')
            tick_label.set_y(top_offset)
        
        tick_label.set_rotation(90)

    # Hide the axes (lines and labels), keep only the tick labels
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)

    # Remove ticks and tick labels from y-axis
    ax.yaxis.set_ticks([])
    ax.set_ylabel('')

    # Adjust layout to ensure labels fit
    plt.tight_layout()



from itertools import product
import dask.array as da
import numpy as np
import xarray as xr
import pylab as plt
import torch
from tqdm import tqdm
from scipy.signal import ricker

import warnings

# Suppress specific warning
warnings.filterwarnings("ignore", message=".*Consider loading the data with Dask directly.*")


def MFP_2D(ds,stations,zrange=[0,100],freqrange=[100,200],xrange=[0,50],vrange=[3000,4000],dz=1,dx=1,dv=100):
    '''
    Perform Matched Field Processing (MFP) to compute beamforming power over a range of 
    spatial coordinates, velocities, and frequencies.

    :param ds: Input data as an xarray DataArray containing waveform data. 
               Must include 'distance' and 'time' coordinates.
    :type ds: xarray.DataArray

    :param zrange: Range of depths to compute (start and end), default is [0, 100].
    :type zrange: list of float or int

    :param freqrange: Frequency range for processing in Hz (start and end), default is [100, 200].
    :type freqrange: list of float

    :param xrange: Range of horizontal x-coordinates to search (start and end), default is [0, 50].
    :type xrange: list of float or int

    :param vrange: Velocity range to search in m/s (start and end), default is [1000, 6000].
    :type vrange: list of float or int

    :param dz: Spacing between grid points in the depth dimension (z), default is 1.
    :type dz: float or int

    :param dx: Spacing between grid points in the horizontal x dimension, default is 1.
    :type dx: float or int

    :param dv: Spacing between velocities, default is 100.
    :type dv: float or int

    :return: A 3D xarray DataArray containing the computed beampower values. The dimensions are
             'v' (velocity), 'z' (depth), and 'x' (horizontal position).
    :rtype: xarray.DataArray
    '''
    
    # Extract data from dataarray
    sampling_rate=((ds.time[1]-ds.time[0]).values.astype(float)*10**-9)
    # Compute FFT of all the receiver
    waveform_spectra=torch.fft.fft(torch.from_numpy(ds.values),axis=1).to(dtype=torch.complex128)

    freqs = torch.fft.fftfreq(len(ds.time),sampling_rate)
    omega = 2 * torch.pi * freqs
    # Build grid for search
    x_coords = torch.arange(xrange[0], xrange[1] + dx, dx)
    z_coords = torch.arange(zrange[0], zrange[1] + dz, dz)
    v_coords = torch.arange(vrange[0], vrange[1] + dv, dv)
    gridpoints = torch.tensor(list(product(x_coords, z_coords)))
    
    stations=torch.tensor(stations).to(dtype=torch.complex128)


    distances_to_all_gridpoints = torch.linalg.norm(gridpoints[:, None, :] - stations[None, :, :], axis=2)
    # COM: This is matching print(distances_to_all_gridpoints)
    # limit to frequency band of interest for
    # a) speed# limit to frequency band of interest for
    # a) speed-up
    # b) focusing on specific frequencies
    freq_idx = torch.where((freqs > freqrange[0]) & (freqs < freqrange[1]))[0]
    omega_lim = omega[freq_idx]
    waveform_spectra_lim = waveform_spectra[:, freq_idx]

    bp=[]
    for vel in tqdm(v_coords):
        

        traveltimes = distances_to_all_gridpoints / vel
        
        # Green's functions between all stations and all grid points
        # within selected frequency band
        # G = exp(-iωt)
        greens_functions = torch.exp(-1j * omega_lim[None, None, :] * traveltimes[:, :, None])
        
        # move critical part to dask
        greens_functions_dask = da.from_array(greens_functions.numpy(), chunks='auto')

        S = (greens_functions_dask[:, :, None, :]*greens_functions_dask.conj()[:, None, :, :])
        
        

        # this assumes that K can be computed without memory issues.
        # If this is not the case, you can also use dask and chunk K.
        K = waveform_spectra_lim[:, None, :] * waveform_spectra_lim.conj()[None, :, :]

        # exclude autocorrelations by filling diagonal with complex zeros
        diag_idxs = torch.arange(K.shape[0])
        zero_spectra = torch.zeros(omega_lim.shape, dtype=torch.cdouble)
        K[diag_idxs, diag_idxs, :] = zero_spectra

        K = da.from_array(K.numpy())

        beampowers_d = da.einsum("gjiw, ijw -> g", S, K).real

        beampowers = beampowers_d.compute()
        bp.append(beampowers.reshape(len(x_coords),len(z_coords)))

    bp=np.stack(bp)
    # bp /= (len(freq_idx)*(len(ds.distance)**2))

    res=xr.DataArray(np.stack(bp),dims=['v','x','z'])
    res['v']=v_coords
    res['x']=x_coords
    res['z']=z_coords

    res=res.transpose("z","x","v")

    return res

def artificial_sources(sensors,sources,velocity,sampling_rate=100,window_length=200,scale=1):
    
    tsensors=torch.tensor(sensors).to(dtype=torch.complex128)
    tsources=torch.tensor(sources).to(dtype=torch.complex128)
    distances = torch.linalg.norm(tsensors - tsources, axis=1)

    traveltimes = (distances / velocity)
    
    # define source wavelet
    times = np.arange(0, window_length + 1 / sampling_rate, 1 / sampling_rate)
    
    # compute frequencies
    freqs = torch.fft.fftfreq(len(times), 1 / sampling_rate)
    omega = 2 * np.pi * freqs

    wavelet = torch.fft.fft(torch.from_numpy(ricker(len(times), scale*sampling_rate)))

    waveform_spectra = wavelet * torch.exp(-1j * omega[None, :] * traveltimes[:, None])
    waveforms = torch.fft.ifft(waveform_spectra, axis=1).real

    da=xr.DataArray(waveforms,dims=['distance','time'])
    da['time']=times*10**9
    return da


def MFP_BH(ds,stations_pos,freqrange=[100,200],xrange=[0,50],yrange=[0,50],zrange=[0,100],vrange=[3000,4000],dx=1,dy=1,dz=1,dv=100):
    '''
    Perform Matched Field Processing (MFP) to compute beamforming power over a range of 
    spatial coordinates, velocities, and frequencies.

    :param ds: Input data as an xarray DataArray containing waveform data. 
               Must include 'distance' and 'time' coordinates.
    :type ds: xarray.DataArray

    :param zrange: Range of depths to compute (start and end), default is [0, 100].
    :type zrange: list of float or int

    :param freqrange: Frequency range for processing in Hz (start and end), default is [100, 200].
    :type freqrange: list of float

    :param xrange: Range of horizontal x-coordinates to search (start and end), default is [0, 50].
    :type xrange: list of float or int

    :param vrange: Velocity range to search in m/s (start and end), default is [1000, 6000].
    :type vrange: list of float or int

    :param dz: Spacing between grid points in the depth dimension (z), default is 1.
    :type dz: float or int

    :param dx: Spacing between grid points in the horizontal x dimension, default is 1.
    :type dx: float or int

    :param dv: Spacing between velocities, default is 100.
    :type dv: float or int

    :return: A 3D xarray DataArray containing the computed beampower values. The dimensions are
             'v' (velocity), 'z' (depth), and 'x' (horizontal position).
    :rtype: xarray.DataArray
    '''
    
    # Extract data from dataarray
    sampling_rate=((ds.time[1]-ds.time[0]).values.astype(float)*10**-9)
    # Compute FFT of all the receiver
    waveform_spectra=torch.fft.fft(torch.from_numpy(ds.values.T),axis=1).to(dtype=torch.complex128)
    freqs = torch.fft.fftfreq(len(ds.time),sampling_rate)
    omega = 2 * torch.pi * freqs
    # Build grid for search
    x_coords = torch.arange(xrange[0], xrange[1] + dx, dx)
    y_coords = torch.arange(yrange[0], yrange[1] + dy, dy)
    z_coords = torch.arange(zrange[0], zrange[1] + dz, dz)
    v_coords = torch.arange(vrange[0], vrange[1] + dv, dv)
    gridpoints = torch.tensor(list(product(x_coords,y_coords,z_coords)))
    
    stations=torch.tensor(stations_pos)

    distances_to_all_gridpoints = torch.linalg.norm(gridpoints[:, None, :] - stations[None, :, :], axis=2)
    # limit to frequency band of interest for
    # a) speed# limit to frequency band of interest for
    # a) speed-up
    # b) focusing on specific frequencies
    freq_idx = torch.where((freqs > freqrange[0]) & (freqs < freqrange[1]))[0]
    omega_lim = omega[freq_idx]
    waveform_spectra_lim = waveform_spectra[:, freq_idx]

    bp=[]
    for vel in tqdm(v_coords):
        
        traveltimes = distances_to_all_gridpoints / vel
        
        # Green's functions between all stations and all grid points
        # within selected frequency band
        # G = exp(-iωt)
        greens_functions = torch.exp(-1j * omega_lim[None, None, :] * traveltimes[:, :, None])
        # move critical part to dask
        greens_functions_dask = da.from_array(greens_functions.numpy(), chunks='auto')

        S = (greens_functions_dask[:, :, None, :]*greens_functions_dask.conj()[:, None, :, :])

        # this assumes that K can be computed without memory issues.
        # If this is not the case, you can also use dask and chunk K.
        K = waveform_spectra_lim[:, None, :] * waveform_spectra_lim.conj()[None, :, :]

        # exclude autocorrelations by filling diagonal with complex zeros
        diag_idxs = torch.arange(K.shape[0])
        zero_spectra = torch.zeros(omega_lim.shape, dtype=torch.cdouble)
        K[diag_idxs, diag_idxs, :] = zero_spectra

        K = da.from_array(K.numpy())

        beampowers_d = da.einsum("gjiw, ijw -> g", S, K).real

        beampowers = beampowers_d.compute()
        bp.append(beampowers.reshape(len(x_coords),len(y_coords),len(z_coords)))

    bp=np.stack(bp)
    bp /= (len(freq_idx)*len(ds.distance))

    res=xr.DataArray(np.stack(bp),dims=['v','x','y','z'])
    res['x']=x_coords
    res['y']=y_coords
    res['z']=z_coords
    res['v']=v_coords

    return res